import numpy as np
from optimask import OptiMask
from sklearn.linear_model import LinearRegression

from ._misc import InvalidEstimatorError, InvalidSubsetError, check_params


class ImputeMultiVariate:
    """
    The `ImputeMultiVariate` class has been developed to address the problem of imputing missing values in multivariate data.
    It relies on regression techniques to estimate missing values using information available in other columns.
    This class offers great flexibility by allowing users to specify a custom regression estimator, while also providing
    a default option to use linear regression from the scikit-learn library. Additionally, it takes into account important parameters
    such as the maximum fraction of missing values allowed in a column and the minimum number of samples required for
    imputation.
    """

    def __init__(self, estimator=None, preprocessing=None, na_frac_max=0.33, min_samples_train=50, weighting_func=None, optimask_n_tries=1, verbose=False):
        self.estimator = self._process_estimator(estimator)
        self.preprocessing = preprocessing
        self.na_frac_max = na_frac_max
        self.min_samples_train = min_samples_train
        self.weighting_func = weighting_func
        self.optimask = OptiMask(n_tries=optimask_n_tries)
        self.verbose = bool(verbose)

    def __repr__(self):
        params = ", ".join(f"{k}={getattr(self, k)}" for k in ('estimator',))
        return f"ImputeMultiVariate({params})"

    @staticmethod
    def _process_estimator(estimator):
        if estimator is None:
            return LinearRegression()
        if not (hasattr(estimator, 'fit') and hasattr(estimator, 'predict')):
            raise InvalidEstimatorError()
        else:
            return estimator

    @staticmethod
    def _process_subset(X, subset, axis):
        n = X.shape[axis]
        check_params(subset, types=(int, list, np.ndarray, tuple, type(None)))
        if isinstance(subset, int):
            if subset >= n:
                raise InvalidSubsetError(f"The subset index {subset} is out of bounds for axis {axis} with size {n}.")
            else:
                return [subset]
        if isinstance(subset, (list, np.ndarray, tuple)):
            return sorted(list(subset))
        if subset is None:
            return list(range(n))

    @staticmethod
    def _prepare_data(mask_nan, col_to_impute, subset_rows):
        rows_to_impute = np.flatnonzero(mask_nan[:, col_to_impute] & ~mask_nan.all(axis=1))
        rows_to_impute = np.intersect1d(ar1=rows_to_impute, ar2=subset_rows)
        other_cols = np.setdiff1d(ar1=np.arange(mask_nan.shape[1]), ar2=[col_to_impute])
        patterns, indexes = np.unique(~mask_nan[np.ix_(rows_to_impute, other_cols)], return_inverse=True, axis=0)
        index_predict = [rows_to_impute[indexes == k] for k in range(len(patterns))]
        columns = [other_cols[pattern] for pattern in patterns]
        return index_predict, columns

    def _prepare_train_and_pred_data(self, X, mask_nan, columns, col_to_impute, index_predict):
        trainable_rows = np.flatnonzero(~mask_nan[:, col_to_impute])
        rows, cols = self.optimask.solve(X[np.ix_(trainable_rows, columns)])
        selected_rows, selected_cols = trainable_rows[rows], columns[cols]
        X_train, y_train = X[np.ix_(selected_rows, selected_cols)], X[selected_rows, col_to_impute]
        X_predict = X[np.ix_(index_predict, selected_cols)]
        return X_train, y_train, X_predict, selected_rows, selected_cols

    def _perform_imputation(self, X_train, y_train, X_predict, selected_rows):
        if callable(self.weighting_func):
            sample_weight = self.weighting_func(selected_rows)
        else:
            sample_weight = None
        model = self.estimator.fit(X_train, y_train, sample_weight=sample_weight)
        return model.predict(X_predict)

    def _impute(self, X, subset_rows, subset_cols):
        ret = np.array(X, dtype=float)
        mask_nan = np.isnan(X)
        imputable_cols = (0 < mask_nan[subset_rows].sum(axis=0)) & (mask_nan.mean(axis=0) <= self.na_frac_max) & (np.nanstd(X, axis=0) > 0)
        imputable_cols = np.intersect1d(np.flatnonzero(imputable_cols), subset_cols)

        for col_to_impute in imputable_cols:
            index_predict, columns = self._prepare_data(mask_nan=mask_nan, col_to_impute=col_to_impute, subset_rows=subset_rows)
            for cols, index in zip(columns, index_predict):
                X_train, y_train, X_predict, selected_rows, selected_cols = self._prepare_train_and_pred_data(X, mask_nan, cols, col_to_impute, index)
                if len(X_train) >= self.min_samples_train:
                    ret[index, col_to_impute] = self._perform_imputation(X_train, y_train, X_predict, selected_rows)
        return ret

    def __call__(self, X, subset_rows=None, subset_cols=None):
        """
        Main method to perform missing value imputation in a multidimensional array.

        Args:
            X (numpy.ndarray): The input array with missing values.
            subset_rows (int, list, numpy.ndarray, tuple, None): The row indices to consider for imputation. If None, all rows are considered. (default: None)
            subset_cols (int, list, numpy.ndarray, tuple, None): The column indices to consider for imputation. If None, all columns are considered. (default: None)

        Returns:
            numpy.ndarray: The array with missing values imputed.
        """
        check_params(X, types=np.ndarray)
        subset_rows = self._process_subset(X=X, subset=subset_rows, axis=0)
        subset_cols = self._process_subset(X=X, subset=subset_cols, axis=1)
        if self.preprocessing is not None:
            Xt = self.preprocessing.fit_transform(X)
        else:
            Xt = X.copy()

        Xt = self._impute(X=Xt, subset_rows=subset_rows, subset_cols=subset_cols)
        if self.preprocessing is not None:
            Xt = self.preprocessing.inverse_transform(Xt)
        return Xt



def check_params(param, params=None, types=None):
    # Check if the parameter's type matches the accepted types
    if (types is not None) and (not isinstance(param, types)):
        if isinstance(types, type):
            accepted = f'{types}'
        else:
            accepted = f"{', '.join([str(t) for t in types])}"
        # Raise a TypeError with a customized message
        msg = f"`{param}` is not of an accepted type, it can only be of type {accepted}!"
        raise TypeError(msg)

    # Check if the parameter is among the recognized parameters
    if (params is not None) and (param not in params):
        # Raise a ValueError with a customized message
        msg = f"`{param}` is not a recognized argument, it can only be one of {', '.join(sorted(params))}!"
        raise ValueError(msg)

    # Return the parameter if it passes the checks
    return param


class InvalidEstimatorError(Exception):
    """Custom exception for invalid estimators."""

    def __init__(self, message="The provided estimator must have both 'fit' and 'predict' methods."):
        self.message = message
        super().__init__(self.message)


class InvalidSubsetError(Exception):
    """Custom exception for invalid subset indices."""

    def __init__(self, message="The provided subset index is out of bounds."):
        self.message = message
        super().__init__(self.message)



# -*- coding: utf-8 -*-

import numpy as np
from sklearn.base import TransformerMixin

__all__ = ['PositiveOutput']


class PositiveOutput(TransformerMixin):
    """
    Parameters:
    - q (float, optional): The quantile used as a threshold for expansion. 
                            Default is q=10, which means the 10th percentile is used as the threshold.
    - v (float, optional): Fixed value used as a threshold for negative expansion.
                            If `v` is specified, this threshold will be used for all features.
                            Default is v=None, which means the threshold is automatically calculated from the data.
    """

    def __init__(self, q=10, v=None):
        if q is None and v is None:
            raise ValueError("At least one of the arguments 'q' or 'v' must be different from None.")

        self.q = q
        self.v = v
        self.thresholds_ = None

    def fit(self, X, y=None):
        """
        Calculate and store the thresholds necessary for negative expansion.

        Parameters:
        - X (array-like): The training data.
        - y (array-like, optional): The training labels. Not used here.

        Returns:
        - self: The fitted PositiveOutput object.
        """
        if np.nanmin(X) < 0:
            raise ValueError("The data must not contain negative values.")

        if self.v is None:
            self.thresholds_ = np.nanpercentile(X, q=self.q, axis=0)
        else:
            self.thresholds_ = np.full(shape=X.shape[1], fill_value=self.v)
        return self

    def transform(self, X, y=None):
        """
        Apply negative expansion on the data.

        Parameters:
        - X (array-like): The data to transform.
        - y (array-like, optional): The labels. Not used here.

        Returns:
        - array-like: The transformed data with negative expansion.
        """
        X = np.asarray(X)
        mask = X < self.thresholds_
        return np.where(mask, 2 * X - self.thresholds_, X)

    def inverse_transform(self, X, y=None):
        """
        Reverse the negative expansion on the transformed data.

        Parameters:
        - X (array-like): The transformed data.
        - y (array-like, optional): The labels. Not used here.

        Returns:
        - array-like: The inverted data after negative expansion.
        """
        X = np.asarray(X)
        mask = X < self.thresholds_
        return np.maximum(0, np.where(mask, 0.5 * X + self.thresholds_ / 2, X))



from ._positive_output_transformer import PositiveOutput
from ._time_series_imputer import ImputeMultiVariate, TimeSeriesImputer

__all__ = ['ImputeMultiVariate', 'PositiveOutput', 'TimeSeriesImputer']



from setuptools import find_packages, setup

setup(
    name='timefiller',
    version='0.1.0',
    packages=find_packages(),
    install_requires=['numpy',
                      'pandas',
                      'optimask',
                      'statsmodels'],
    python_requires='>=3.8',
    author='Cyril Joly',
    description='A package for imputing missing data in time series',
    classifiers=['License :: OSI Approved :: MIT License'],
)



import numpy as np
import pandas as pd
import statsmodels.api as sm
from sklearn.feature_selection import r_regression
from tqdm.auto import tqdm

from ._misc import check_params
from ._multivariate_imputer import ImputeMultiVariate


class TimeSeriesImputer:
    """Class for time series imputation.

    Args:
        estimator (object, optional): Estimator used for imputation.
        preprocessing (callable, optional): Data preprocessing.
        ar_lags (int, list, numpy.ndarray, or tuple, optional): Autoregressive lags to consider.
        multivariate_lags (int or None, optional): Multivariate lags to consider.
        na_frac_max (float, optional): Maximum fraction of missing values allowed.
        min_samples_train (int, optional): Minimum number of samples for training.
        weighting_func (callable, optional): Weighting function for imputation.
        optimask_n_tries (int, optional): Number of attempts for optimization.
        verbose (bool, optional): Display process details.
        random_state (int or None, optional): Random state for reproducibility.
    """

    def __init__(self, estimator=None, preprocessing=None, ar_lags=None, multivariate_lags=None, na_frac_max=0.33,
                 min_samples_train=50, weighting_func=None, optimask_n_tries=1, verbose=False, random_state=None):

        self.imputer = ImputeMultiVariate(estimator=estimator, preprocessing=preprocessing,
                                          na_frac_max=na_frac_max, min_samples_train=min_samples_train,
                                          weighting_func=weighting_func, optimask_n_tries=optimask_n_tries,
                                          verbose=verbose)
        self.ar_lags = self._process_lags(ar_lags)
        self.multivariate_lags = check_params(multivariate_lags, types=(int, type(None)))
        self.verbose = bool(verbose)
        self.random_state = random_state

    def __repr__(self):
        params = ", ".join(f"{k}={getattr(self, k)}" for k in ('ar_lags', 'multivariate_lags'))
        return f"TimeSeriesImputer({params})"

    @staticmethod
    def _process_lags(ar_lags):
        check_params(ar_lags, types=(int, list, np.ndarray, tuple, type(None)))
        if ar_lags is None:
            return None
        if isinstance(ar_lags, int):
            ar_lags = list(range(-abs(ar_lags)-1, -1)) + list(range(1, abs(ar_lags)+1))
            return sorted(ar_lags)
        if isinstance(ar_lags, (tuple, list, np.ndarray)):
            ar_lags = [-k for k in ar_lags if k != 0] + [k for k in ar_lags if k != 0]
            return sorted(list(set(ar_lags)))

    @staticmethod
    def _sample_features(data, col, n_nearest_features, rng):
        x = data.fillna(data.mean())
        # computes pearson correlation between col and others series
        s1 = r_regression(X=x.drop(columns=col), y=x[col])
        # computes the mean number of timestamps containing common valid data between col and others series
        s2 = ((~data[col].isnull()).astype(float).values@(~data.drop(columns=col).isnull()).astype(float).values)/len(data)
        # features are sampled according those computed features
        p = np.sqrt(abs(s1) * s2)
        size = min(n_nearest_features, len(s1), len(p[p > 0]))
        cols_to_sample = list(data.drop(columns=col).columns)
        return list(rng.choice(a=cols_to_sample, size=size, p=p/p.sum(), replace=False))

    @staticmethod
    def _best_lag(s1, s2, max_lags):
        c1 = sm.tsa.ccf(s1, s2, nlags=max_lags)[::-1]
        c2 = sm.tsa.ccf(s2, s1, nlags=max_lags)[1:]
        c = np.concatenate([c1, c2])
        return np.abs(c).argmax() - max_lags + 1

    @classmethod
    def find_best_lags(cls, x, col, max_lags):
        df = x.fillna(x.mean())
        cols = df.drop(columns=col).columns
        ret = [x[col]]
        for other_col in cols:
            lag = cls._best_lag(df[col], df[other_col], max_lags=max_lags)
            if lag != 0:
                ret.append(x[other_col].shift(-lag).rename(f"{other_col}{-lag:+d}"))
            else:
                ret.append(x[other_col])
        return pd.concat(ret, axis=1)

    @staticmethod
    def _process_subset_cols(X, subset_cols):
        _, n = X.shape
        columns = list(X.columns)
        if subset_cols is None:
            return list(range(n))
        if isinstance(subset_cols, str):
            if subset_cols in columns:
                return [columns.index(subset_cols)]
            else:
                return []
        if isinstance(subset_cols, (list, tuple, pd.core.indexes.base.Index)):
            return [columns.index(_) for _ in subset_cols if _ in columns]
        raise TypeError(f"subset_cols should be of type str, list, tuple, or pandas Index. Received type {type(subset_cols)} instead.")

    @staticmethod
    def _process_subset_rows(X, before, after):
        index = pd.Series(np.arange(len(X)), index=X.index)
        if before is not None:
            index = index[index.index <= pd.to_datetime(str(before))]
        if after is not None:
            index = index[pd.to_datetime(str(after)) <= index.index]
        return list(index.values)

    def _impute_col(self, x, col, subset_rows):
        if isinstance(self.multivariate_lags, int):
            x = self.find_best_lags(x, col, self.multivariate_lags)
        x = x.copy()
        if self.ar_lags is not None:
            for k in sorted(self.ar_lags):
                x[f"{col}{k:+d}"] = x[col].shift(k).copy()
        index_col = list(x.columns).index(col)
        x_col_imputed = self.imputer(x.values, subset_rows=subset_rows, subset_cols=index_col)[:, index_col]
        return pd.Series(x_col_imputed, name=col, index=x.index)

    def __call__(self, X, subset_cols=None, before=None, after=None, n_nearest_features=None) -> pd.DataFrame:
        """Call method for imputation.

        Args:
            X (DataFrame): Data to be imputed.
            subset_cols (str, list, tuple, or pandas.core.indexes.base.Index, optional): Columns to be imputed. By default, all columns will be imputed.
            before (str or pd.Timestamp or None, optional): Date before which the data is imputed. By default, no lower temporal limit is set.
            after (str or pd.Timestamp or None, optional): Date after which the data is imputed. By default, no upper temporal limit is set.
            n_nearest_features (int, optional): Number of nearest features to consider. A heuristic is used: the features are selected randomly, based on
            their correlations with the feature to be imputed, as well as the number of common temporal observations with the feature to be imputed.

        Returns:
            DataFrame or tuple: Imputed data.
        """

        rng = np.random.default_rng(self.random_state)
        X_ = check_params(X, types=pd.DataFrame).copy()
        check_params(X_.index, types=pd.DatetimeIndex)

        if X_.index.freq is None:
            X_ = X_.asfreq(pd.infer_freq(X_.index))
        X_ = X_[X_.columns[X_.std() > 0]].copy()
        columns = list(X_.columns)

        ret = [pd.Series(index=X.index)]
        subset_rows = self._process_subset_rows(X_, before, after)
        subset_cols = self._process_subset_cols(X_, subset_cols)
        for index_col in tqdm(subset_cols, disable=(not self.verbose)):
            col = columns[index_col]
            if isinstance(n_nearest_features, int):
                cols_in = [col] + self._sample_features(X_, col, n_nearest_features, rng)
            else:
                cols_in = list(X_.columns)
            ret.append(self._impute_col(x=X_[cols_in], col=col, subset_rows=subset_rows))
        ret = pd.concat(ret, axis=1).reindex_like(X).combine_first(X)
        return ret
